-- Total RP 3 Nameplate Module
-- Copyright 2019 Total RP 3 Development Team
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
local _, TRP3_API = ...;

-- TRP3_API imports.
local L = TRP3_API.loc;
local TRP3_Events = TRP3_API.events;
local TRP3_Globals = TRP3_API.globals;

-- Nameplate module table.
local NamePlates = {
	-- Active decorator that is used to customize nameplates.
	decorator = nil,
};

-- Updates a single nameplate for the given unit token, if one exists.
--
-- Return true if a nameplate is successfully updated, or false if not.
function NamePlates.UpdateNamePlateForUnit(unitToken)
	-- Validate the unit token.
	if not NamePlates.IsUnitValid(unitToken) then
		return false;
	end

	-- Grab the decorator if one exists and trigger the update.
	local decorator = NamePlates.GetCurrentDecorator();
	if not decorator then
		return false;
	end

	return decorator:UpdateNamePlateForUnit(unitToken);
end

-- Updates all nameplates managed by this module.
function NamePlates.UpdateAllNamePlates()
	local decorator = NamePlates.GetCurrentDecorator();
	if not decorator then
		return;
	end

	return decorator:UpdateAllNamePlates();
end

-- Returns the mixin that should be used to provide a nameplate display.
--[[private]] function NamePlates.GetSuggestedDecorator()
	-- Add any supported addons here.
	if IsAddOnLoaded("Kui_Nameplates") then
		return NamePlates.KuiDecoratorMixin;
	end

	-- ElvUI has a separately configurable module for its nameplates;
	-- users sometimes opt to use another addon instead, so we'll check
	-- for ElvUI near the end.
	local aceAddon = LibStub:GetLibrary("AceAddon-3.0", true);
	local addonElvUI = aceAddon and aceAddon:GetAddon("ElvUI", true);
	if addonElvUI then
		-- Helpfully, ElvUI's nameplate module may not be initialized at
		-- this point so the only option we realistically have is this.
		local private = addonElvUI.private;
		if private and private.nameplates and private.nameplates.enable then
			return NamePlates.ElvUIDecoratorMixin;
		end
	end

	-- Last resort; Blizzard defaults. Don't put addons below here.
	if IsAddOnLoaded("Blizzard_NamePlates") then
		return NamePlates.BlizzardDecoratorMixin;
	end

	-- Nothing is supported.
	return nil;
end

-- Returns the decorator in use for the nameplate display.
--[[private]] function NamePlates.GetCurrentDecorator()
	return NamePlates.decorator;
end

-- Sets the decorator to use for the nameplate display.
--[[private]] function NamePlates.SetCurrentDecorator(decorator)
	NamePlates.decorator = decorator;
end

-- Handler triggered then the game assigns a nameplate unit token.
--[[private]] function NamePlates.OnNamePlateUnitAdded(unitToken)
	-- Issue a request for the profile ahead of notifying the decorator.
	if NamePlates.ShouldRequestUnitProfile(unitToken) then
		NamePlates.RequestUnitProfile(unitToken);
	end
end

-- Handler triggered when the player leaves combat.
--[[private]] function NamePlates.OnPlayerRegenEnabled()
	-- Update our customization state for all nameplates.
	NamePlates.SetInCombat(false);
	NamePlates.UpdateCustomizationState();
end

-- Handler triggered when the player enters combat.
--[[private]] function NamePlates.OnPlayerRegenDisabled()
	-- Update our customization state for all nameplates.
	NamePlates.SetInCombat(true);
	NamePlates.UpdateCustomizationState();
end

-- Handler triggered when a unit's name updates.
--[[private]] function NamePlates.OnUnitNameUpdate(unitToken)
	-- Reject any non-nameplate units.
	if not strfind(unitToken, "nameplate", 1, true) then
		return;
	end

	-- Issue a request for the profile.
	if NamePlates.ShouldRequestUnitProfile(unitToken) then
		NamePlates.RequestUnitProfile(unitToken);
	end
end

-- Handler triggered when a configuration setting is changed.
--[[private]] function NamePlates.OnConfigSettingChanged(key, _)
	-- Only nameplate setting changes and color contrast updates, please.
	if not strfind(tostring(key), "^nameplates_") and key ~= "increase_color_contrast" then
		return;
	end

	NamePlates.UpdateCustomizationState();
end

-- Handler triggered when the player mouses over an in-game unit.
--[[private]] function NamePlates.OnMouseOverChanged()
	-- The nameplate API can link mouseover/target/etc. units to frames, so
	-- just validate it.
	if not NamePlates.IsUnitValid("mouseover") then
		return;
	end

	NamePlates.UpdateNamePlateForUnit("mouseover");
end

-- Handler triggered when TRP updates the registry for a named profile.
--[[private]] function NamePlates.OnRegisterDataUpdated(registerID)
	-- Get the unit token associated with the updated profile.
	local unitToken = NamePlates.GetUnitForRegisterID(registerID);
	if not NamePlates.IsUnitValid(unitToken) then
		return;
	end

	NamePlates.UpdateNamePlateForUnit(unitToken);
end

-- Handler triggered when the roleplay status of the character changes, such
-- as from IC to OOC.
--[[private]] function NamePlates.OnRoleplayStatusChanged()
	-- Update internal state and then all the nameplates.
	NamePlates.UpdateCustomizationState();
	NamePlates.UpdateAllNamePlates();
end

-- Module lifecycle functions.

-- Handler called when the module is being initialized. This is responsible
-- for setting up the base internal state without actually enabling any
-- behaviours.
--[[private]] function NamePlates.OnModuleInitialize()
	-- Register the configuration keys early so we can access them whenever.
	NamePlates.RegisterConfigurationKeys();

	-- Prevent the module from actually being set up further if any
	-- known addon conflicts exists.
	for _, addonName in ipairs(NamePlates.CONFLICTING_ADDONS) do
		local state = GetAddOnEnableState(TRP3_Globals.player, addonName);
		if state == 2 then
			return false, format(L.NAMEPLATES_ERR_ADDON_CONFLICT, addonName);
		end
	end
end

-- Handler called when the module is started up. This is responsible for
-- fully starting the module, registering events and hooks as needed.
--[[private]] function NamePlates.OnModuleStart()
	-- Activate an appropriate decorator based on the environment.
	local decoratorMixin = NamePlates.GetSuggestedDecorator();
	if not decoratorMixin then
		return false, L.NAMEPLATES_ERR_NO_VALID_PROVIDER;
	end

	-- Initialize and activate it.
	local decorator = CreateAndInitFromMixin(decoratorMixin);
	NamePlates.SetCurrentDecorator(decorator);

	-- Register events and script handlers.
	local eventFrame = CreateFrame("Frame");
	eventFrame:RegisterEvent("NAME_PLATE_UNIT_ADDED");
	eventFrame:RegisterEvent("PLAYER_REGEN_ENABLED");
	eventFrame:RegisterEvent("PLAYER_REGEN_DISABLED");
	eventFrame:RegisterEvent("UNIT_NAME_UPDATE");

	eventFrame:SetScript("OnEvent", function(_, event, ...)
		if event == "NAME_PLATE_UNIT_ADDED" then
			NamePlates.OnNamePlateUnitAdded(...);
		elseif event == "PLAYER_REGEN_ENABLED" then
			NamePlates.OnPlayerRegenEnabled(...);
		elseif event == "PLAYER_REGEN_DISABLED" then
			NamePlates.OnPlayerRegenDisabled(...);
		elseif event == "UNIT_NAME_UPDATE" then
			NamePlates.OnUnitNameUpdate(...);
		end
	end);

	-- Register internal events.
	TRP3_Events.registerCallback("MOUSE_OVER_CHANGED", NamePlates.OnMouseOverChanged);
	TRP3_Events.registerCallback("REGISTER_DATA_UPDATED", NamePlates.OnRegisterDataUpdated);
	TRP3_Events.registerCallback("CONFIG_SETTING_CHANGED", NamePlates.OnConfigSettingChanged);
	TRP3_Events.registerCallback("ROLEPLAY_STATUS_CHANGED", NamePlates.OnRoleplayStatusChanged);

	-- Set up a timer to periodically prune the cooldown table.
	C_Timer.NewTicker(30, NamePlates.PruneUnitRequestCooldowns);

	-- Install the configuration UI.
	NamePlates.RegisterConfigurationUI();

	-- Update the initial state of our customizations.
	NamePlates.UpdateCustomizationState();

	-- Trigger an update of all nameplates.
	NamePlates.UpdateAllNamePlates();
end

-- Module exports.
AddOn_TotalRP3.NamePlates = NamePlates;

-- Module registration.
TRP3_API.module.registerModule({
	name        = L.NAMEPLATES_MODULE_NAME,
	description = L.NAMEPLATES_MODULE_DESCRIPTION,
	version     = 1,
	id          = "trp3_nameplates",
	onInit      = NamePlates.OnModuleInitialize,
	onStart     = NamePlates.OnModuleStart,
	minVersion  = 70,
});
